@@ -1,5 +1,9 @@
+# SPDX-License-Identifier: MIT
+
 """Unit tests for mbedtls.random."""
 
+from __future__ import annotations
+
 import pickle
 import random
 from collections import defaultdict
@@ -8,7 +12,7 @@ from typing import Callable, MutableMapping, Sequence
 import pytest
 
 import mbedtls._random as _drbg
-import mbedtls.mpi as _mpi  # type: ignore
+import mbedtls.mpi as _mpi
 from mbedtls.exceptions import TLSError
 
 
