@@ -1,6 +1,9 @@
+# SPDX-License-Identifier: MIT
+
 from __future__ import annotations
 
 import datetime as dt
+import errno
 import pickle
 import socket
 import subprocess
@@ -356,6 +359,24 @@ class TestDTLSCookie:
         assert cookie.timeout == 1000
 
 
+def assert_conf_invariant(
+    conf: Union[TLSConfiguration, DTLSConfiguration],
+    **elem: Any,
+) -> None:
+    # For the context, the configuration is converted into an
+    # internal MbedTLSConfiguration and the accessors then
+    # converts the MbedTLSConfiguration back into a
+    # TLSConfiguration (or DTLSConfiguration).  These conversions
+    # are what we actually check here.
+    __tracebackhide__ = True
+    assert len(elem) == 1
+
+    key, value = next(iter(elem.items()))
+    assert (
+        getattr(ServerContext(conf.update(**elem)).configuration, key) == value
+    )
+
+
 class TestConfiguration:
     @pytest.fixture(params=[DTLSConfiguration, TLSConfiguration])
     def conf(self, request: Any) -> Union[TLSConfiguration, DTLSConfiguration]:
@@ -365,66 +386,56 @@ class TestConfiguration:
         )
         return conf
 
-    @pytest.mark.parametrize(
-        "repr_",
-        [repr, str],
-        ids=lambda f: f.__name__,  # type: ignore[no-any-return]
-    )
-    def test_repr(
-        self,
-        repr_: Callable[[object], str],
-        conf: Union[TLSConfiguration, DTLSConfiguration],
-    ) -> None:
-        assert isinstance(repr_(conf), str)
-
     @pytest.mark.parametrize("validate", [True, False])
     def test_set_validate_certificates(
         self, conf: Union[TLSConfiguration, DTLSConfiguration], validate: bool
     ) -> None:
-        conf_ = conf.update(validate_certificates=validate)
-        assert conf_.validate_certificates is validate
+        assert_conf_invariant(conf, validate_certificates=validate)
+
+    def test_set_empty_certificate_chain(
+        self, conf: Union[TLSConfiguration, DTLSConfiguration]
+    ) -> None:
+        assert_conf_invariant(conf, certificate_chain=None)
 
-    @pytest.mark.parametrize("chain", [((), None), None])
     def test_set_certificate_chain(
-        self,
-        conf: Union[TLSConfiguration, DTLSConfiguration],
-        chain: Tuple[Tuple[CRT, ...], _Key],
+        self, conf: Union[TLSConfiguration, DTLSConfiguration]
     ) -> None:
         root_crt, root_key = make_root_ca()
         ee_crt, ee_key = make_crt(root_crt, root_key)
-        if chain is None:
-            chain = (ee_crt, root_crt), ee_key
-        conf_ = conf.update(certificate_chain=chain)
-        assert conf_.certificate_chain == chain
+        chain = (ee_crt, root_crt), ee_key
+        assert_conf_invariant(conf, certificate_chain=chain)
 
     def test_set_ciphers(
         self, conf: Union[TLSConfiguration, DTLSConfiguration]
     ) -> None:
-        ciphers = tuple(ciphers_available())
-        conf_ = conf.update(ciphers=ciphers)
-        assert conf_.ciphers == ciphers
+        assert_conf_invariant(conf, ciphers=tuple(ciphers_available()))
 
     @pytest.mark.parametrize(
         "inner_protocols",
-        [(), (NextProtocol.H2, NextProtocol.H2C), (b"h2", b"h2c", b"ftp")],
+        [
+            (),
+            (NextProtocol.H2, NextProtocol.H2C),
+            (NextProtocol.H2, NextProtocol.H2C, NextProtocol.FTP),
+        ],
     )
     def test_set_inner_protocols(
         self,
         conf: Union[TLSConfiguration, DTLSConfiguration],
         inner_protocols: Tuple[Union[NextProtocol, bytes]],
     ) -> None:
-        conf_ = conf.update(inner_protocols=inner_protocols)
-        assert conf_.inner_protocols == inner_protocols
+        assert_conf_invariant(
+            conf,
+            inner_protocols=inner_protocols,
+        )
 
-    @pytest.mark.parametrize("store", [TrustStore.system()])
+    @pytest.mark.parametrize("trust_store", [TrustStore.system()])
     def test_trust_store(
         self,
         conf: Union[TLSConfiguration, DTLSConfiguration],
-        store: TrustStore,
+        trust_store: TrustStore,
     ) -> None:
-        conf_ = conf.update(trust_store=store)
-        assert store
-        assert conf_.trust_store == store
+        assert trust_store
+        assert_conf_invariant(conf, trust_store=trust_store)
 
     @pytest.mark.parametrize("callback", [None])
     def test_set_sni_callback(
@@ -441,8 +452,7 @@ class TestConfiguration:
         psk: Tuple[str, bytes],
     ) -> None:
         assert conf.pre_shared_key is None
-        conf_ = conf.update(pre_shared_key=psk)
-        assert conf_.pre_shared_key == psk
+        assert_conf_invariant(conf, pre_shared_key=psk)
 
     @pytest.mark.parametrize("psk_store", [{}, {"client": b"the secret key"}])
     def test_psk_store(
@@ -451,8 +461,7 @@ class TestConfiguration:
         psk_store: Mapping[str, bytes],
     ) -> None:
         assert not conf.pre_shared_key_store
-        conf_ = conf.update(pre_shared_key_store=psk_store)
-        assert conf_.pre_shared_key_store == psk_store
+        assert_conf_invariant(conf, pre_shared_key_store=psk_store)
 
 
 class TestTLSConfiguration:
@@ -468,15 +477,13 @@ class TestTLSConfiguration:
     def test_lowest_supported_version(
         self, conf: TLSConfiguration, version: TLSVersion
     ) -> None:
-        conf_ = conf.update(lowest_supported_version=version)
-        assert conf_.lowest_supported_version is version
+        assert_conf_invariant(conf, lowest_supported_version=version)
 
     @pytest.mark.parametrize("version", _SUPPORTED_TLS_VERSION)
     def test_highest_supported_version(
         self, conf: TLSConfiguration, version: TLSVersion
     ) -> None:
-        conf_ = conf.update(highest_supported_version=version)
-        assert conf_.highest_supported_version is version
+        assert_conf_invariant(conf, highest_supported_version=version)
 
 
 class TestDTLSConfiguration:
@@ -492,23 +499,20 @@ class TestDTLSConfiguration:
     def test_lowest_supported_version(
         self, conf: DTLSConfiguration, version: DTLSVersion
     ) -> None:
-        conf_ = conf.update(lowest_supported_version=version)
-        assert conf_.lowest_supported_version is version
+        assert_conf_invariant(conf, lowest_supported_version=version)
 
     @pytest.mark.parametrize("version", _SUPPORTED_DTLS_VERSION)
     def test_highest_supported_version(
         self, conf: DTLSConfiguration, version: DTLSVersion
     ) -> None:
-        conf_ = conf.update(highest_supported_version=version)
-        assert conf_.highest_supported_version is version
+        assert_conf_invariant(conf, highest_supported_version=version)
 
     @pytest.mark.parametrize("anti_replay", [True, False])
     def test_set_anti_replay(
         self, conf: DTLSConfiguration, anti_replay: bool
     ) -> None:
         assert conf.anti_replay is True
-        conf_ = conf.update(anti_replay=anti_replay)
-        assert conf_.anti_replay is anti_replay
+        assert_conf_invariant(conf, anti_replay=anti_replay)
 
     @pytest.mark.parametrize(
         "hs_min_max", [(1, 60), (42, 69), (4.2, 6.9), (42.0, 69.0)]
@@ -528,6 +532,9 @@ class TestDTLSConfiguration:
         assert conf_.handshake_timeout_min == hs_min
         assert conf_.handshake_timeout_max == hs_max
 
+        assert_conf_invariant(conf, handshake_timeout_min=hs_min)
+        assert_conf_invariant(conf, handshake_timeout_max=hs_max)
+
     @pytest.mark.parametrize("hs_min_max", [(1, 60), (42, 69), (4.2, 6.9)])
     def test_handshake_timeout_default(
         self,
@@ -545,6 +552,9 @@ class TestDTLSConfiguration:
         assert conf_.handshake_timeout_min == hs_min
         assert conf_.handshake_timeout_max == hs_max
 
+        assert_conf_invariant(conf, handshake_timeout_min=hs_min)
+        assert_conf_invariant(conf, handshake_timeout_max=hs_max)
+
 
 class TestContext:
     @pytest.fixture(params=[Purpose.SERVER_AUTH, Purpose.CLIENT_AUTH])
@@ -605,19 +615,6 @@ HANDSHAKE_OVER = (
 )
 
 
-def make_server(
-    conf: Union[TLSConfiguration, DTLSConfiguration]
-) -> TLSWrappedBuffer:
-    return ServerContext(conf).wrap_buffers()
-
-
-def make_client(
-    conf: Union[TLSConfiguration, DTLSConfiguration],
-    hostname: Optional[_HostName],
-) -> TLSWrappedBuffer:
-    return ClientContext(conf).wrap_buffers(hostname)
-
-
 def do_io(
     *, src: TLSWrappedBuffer, dst: TLSWrappedBuffer, amt: int = 1024
 ) -> None:
@@ -711,10 +708,10 @@ def do_communicate(args: Any) -> str:
             encoding="utf8",
         ) as proc:
             out, err = proc.communicate()
-            if "ConnectionRefusedError" not in err:
-                return out
-            time.sleep(0.01)  # Avoid tight CPU loop.
-            continue
+            if "ConnectionRefusedError" in err:
+                time.sleep(0.01)  # Avoid tight CPU loop.
+                continue
+            return out
 
 
 class TestTLSHandshake:
@@ -735,18 +732,18 @@ class TestTLSHandshake:
     def test_cert_without_validation(
         self, certificate_chain: Tuple[Tuple[CRT, ...], _Key]
     ) -> None:
-        server = make_server(
+        server = ServerContext(
             TLSConfiguration(
                 certificate_chain=certificate_chain,
                 validate_certificates=False,
             )
-        )
-        client = make_client(
-            TLSConfiguration(validate_certificates=False), "hostname"
-        )
+        ).wrap_buffers()
+        client = ClientContext(
+            TLSConfiguration(validate_certificates=False)
+        ).wrap_buffers("hostname")
         make_full_handshake(client=client, server=server)
 
-        secret = "a very secret message".encode("utf8")
+        secret = b"a very secret message"
         assert do_send(secret, src=client, dst=server) == secret
         assert do_send(secret, src=server, dst=client) == secret
 
@@ -759,40 +756,39 @@ class TestTLSHandshake:
         crt: CRT
         for crt in certificate_chain[0][1:]:
             trust_store.add(crt)
-        server = make_server(
+        server = ServerContext(
             TLSConfiguration(
                 certificate_chain=certificate_chain,
                 validate_certificates=False,
             )
-        )
+        ).wrap_buffers()
         # Host name must now be the common name (CN) of the leaf certificate.
-        client = make_client(
-            TLSConfiguration(trust_store=trust_store), hostname
-        )
+        client = ClientContext(
+            TLSConfiguration(trust_store=trust_store)
+        ).wrap_buffers(hostname)
         make_full_handshake(client=client, server=server)
 
-        secret = "a very secret message".encode("utf8")
+        secret = b"a very secret message"
         assert do_send(secret, src=client, dst=server) == secret
         assert do_send(secret, src=server, dst=client) == secret
 
     def test_psk(self) -> None:
         psk = ("cli", b"secret")
-        server = make_server(
+        server = ServerContext(
             TLSConfiguration(
                 pre_shared_key_store=dict((psk,)),
                 validate_certificates=False,
             )
-        )
-        client = make_client(
+        ).wrap_buffers()
+        client = ClientContext(
             TLSConfiguration(
                 pre_shared_key=psk,
                 validate_certificates=False,
             ),
-            "hostname",
-        )
+        ).wrap_buffers("hostname")
         make_full_handshake(client=client, server=server)
 
-        secret = "a very secret message".encode("utf8")
+        secret = b"a very secret message"
         assert do_send(secret, src=client, dst=server) == secret
         assert do_send(secret, src=server, dst=client) == secret
 
@@ -800,51 +796,49 @@ class TestTLSHandshake:
 class TestDTLSHandshake:
     def test_psk(self) -> None:
         psk = ("cli", b"secret")
-        server = make_server(
+        server = ServerContext(
             DTLSConfiguration(
                 pre_shared_key_store=dict((psk,)),
                 validate_certificates=False,
             )
-        )
-        client = make_client(
+        ).wrap_buffers()
+        client = ClientContext(
             DTLSConfiguration(
                 pre_shared_key=psk,
                 validate_certificates=False,
             ),
-            "hostname",
-        )
+        ).wrap_buffers("hostname")
         make_hello_verify_request(
-            client=client, server=server, cookie="🍪🍪🍪".encode("utf-8")
+            client=client, server=server, cookie="🍪🍪🍪".encode()
         )
         make_full_handshake(client=client, server=server)
 
-        secret = "a very secret message".encode("utf8")
+        secret = b"a very secret message"
         assert do_send(secret, src=client, dst=server) == secret
         assert do_send(secret, src=server, dst=client) == secret
 
     def test_resume_from_pickle(self) -> None:
         psk = ("cli", b"secret")
-        server = make_server(
+        server = ServerContext(
             DTLSConfiguration(
                 pre_shared_key_store=dict((psk,)),
                 validate_certificates=False,
             )
-        )
-        client = make_client(
+        ).wrap_buffers()
+        client = ClientContext(
             DTLSConfiguration(
                 pre_shared_key=psk,
                 ciphers=("TLS-ECDHE-PSK-WITH-CHACHA20-POLY1305-SHA256",),
                 lowest_supported_version=DTLSVersion.DTLSv1_2,
                 validate_certificates=False,
             ),
-            "hostname",
-        )
+        ).wrap_buffers("hostname")
         make_hello_verify_request(
-            client=client, server=server, cookie="🍪🍪🍪".encode("utf-8")
+            client=client, server=server, cookie="🍪🍪🍪".encode()
         )
         make_full_handshake(client=client, server=server)
 
-        secret = "a very secret message".encode("utf8")
+        secret = b"a very secret message"
         do_send(secret, src=client, dst=server)
         do_send(secret, src=server, dst=client)
 
@@ -855,7 +849,7 @@ class TestDTLSHandshake:
 
 @pytest.mark.skipif(sys.platform == "win32", reason="Flaky under Windows")
 class TestProgramsTLS:
-    @pytest.fixture()
+    @pytest.fixture(scope="class")
     def port(self) -> int:
         """Return a free port
 
@@ -869,7 +863,7 @@ class TestProgramsTLS:
             port = sock.getsockname()[1]
         return port
 
-    @pytest.fixture()
+    @pytest.fixture(scope="class")
     def server(
         self, rootpath: Path, port: int
     ) -> Iterator[subprocess.Popen[str]]:
@@ -882,14 +876,15 @@ class TestProgramsTLS:
             "--psk-store",
             "cli=secret",
         ]
-        proc = subprocess.Popen(args, text=True, encoding="utf8")
-        yield proc
-        proc.kill()
-        proc.wait(1.0)
+        with subprocess.Popen(args, text=True, encoding="utf8") as proc:
+            yield proc
+            proc.kill()
+            proc.wait(1.0)
 
+    @pytest.mark.repeat(3)
     @pytest.mark.usefixtures("server")
-    @pytest.mark.timeout(10)
-    def test_communicate(self, rootpath: Path, port: int) -> None:
+    @pytest.mark.timeout(2)
+    def test_client(self, rootpath: Path, port: int) -> None:
         secret = "a very secret message"
         args = [
             sys.executable,
@@ -904,10 +899,86 @@ class TestProgramsTLS:
         for _ in range(3):
             assert do_communicate(args) == secret + "\n"
 
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_send_recv(self, port: int) -> None:
+        secret = b"a very secret message"
+        with ClientContext(
+            TLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_STREAM), "localhost"
+        ) as client:
+            client.connect(("127.0.0.1", port))
+            client.do_handshake()
+            sent = client.send(secret)
+            assert sent == len(secret)
+
+            data = client.recv(1024)
+        assert data == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_send_recv_into(self, port: int) -> None:
+        secret = b"a very secret message"
+        buffer = bytearray(b"\0" * 256)
+        with ClientContext(
+            TLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_STREAM), "localhost"
+        ) as client:
+            client.connect(("127.0.0.1", port))
+            client.do_handshake()
+            sent = client.send(secret)
+            assert sent == len(secret)
+
+            received = client.recv_into(buffer, 1024)
+        assert buffer[:received] == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_sendall_recv(self, port: int) -> None:
+        secret = b"a very secret message"
+        with ClientContext(
+            TLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_STREAM), "localhost"
+        ) as client:
+            client.connect(("127.0.0.1", port))
+            client.do_handshake()
+            client.sendall(secret)
+            data = client.recv(1024)
+        assert data == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_connectionless_unavailable(self, port: int) -> None:
+        address = ("127.0.0.1", port)
+        with ClientContext(
+            TLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_STREAM), "localhost"
+        ) as client:
+            with pytest.raises(OSError) as excinfo:
+                client.do_handshake(address)
+
+        assert excinfo.value.errno is errno.ENOTCONN
+
 
 @pytest.mark.skipif(sys.platform == "win32", reason="Flaky under Windows")
 class TestProgramsDTLS:
-    @pytest.fixture()
+    @pytest.fixture(scope="class")
     def port(self) -> int:
         """Return a free port
 
@@ -921,7 +992,7 @@ class TestProgramsDTLS:
             port = sock.getsockname()[1]
         return port
 
-    @pytest.fixture()
+    @pytest.fixture(scope="class")
     def server(
         self, rootpath: Path, port: int
     ) -> Iterator[subprocess.Popen[str]]:
@@ -939,9 +1010,10 @@ class TestProgramsDTLS:
         proc.kill()
         proc.wait(1.0)
 
+    @pytest.mark.timeout(3)
     @pytest.mark.usefixtures("server")
-    @pytest.mark.timeout(10)
-    def test_communication(self, rootpath: Path, port: int) -> None:
+    @pytest.mark.timeout(2)
+    def test_client(self, rootpath: Path, port: int) -> None:
         secret = "a very secret message"
         args = [
             sys.executable,
@@ -955,3 +1027,90 @@ class TestProgramsDTLS:
         ]
         for _ in range(3):
             assert do_communicate(args) == secret + "\n"
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_send_recv(self, port: int) -> None:
+        address = ("127.0.0.1", port)
+        secret = b"a very secret message"
+        with ClientContext(
+            DTLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_DGRAM), "localhost"
+        ) as client:
+            client.connect(address)
+            client.do_handshake()
+            sent = client.send(secret)
+            assert sent == len(secret)
+
+            data = client.recv(1024)
+        assert data == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_sendto_recvfrom(self, port: int) -> None:
+        address = ("127.0.0.1", port)
+        secret = b"a very secret message"
+        with ClientContext(
+            DTLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_DGRAM), "localhost"
+        ) as client:
+            client.do_handshake(address)
+            sent = client.sendto(secret, address)
+            assert sent == len(secret)
+
+            data, addr = client.recvfrom(1024)
+            assert addr == address
+        assert data == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_sendto_recvfrom_with_flags(self, port: int) -> None:
+        # Note that flags is always 0 (noop) here because we are only
+        # interested in testing the API.  See also issue #62.
+        address = ("127.0.0.1", port)
+        secret = b"a very secret message"
+        with ClientContext(
+            DTLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_DGRAM), "localhost"
+        ) as client:
+            client.do_handshake(0, address)
+            sent = client.sendto(secret, 0, address)
+            assert sent == len(secret)
+
+            data, addr = client.recvfrom(1024, 0)
+            assert addr == address
+        assert data == secret
+
+    @pytest.mark.usefixtures("server")
+    @pytest.mark.timeout(2)
+    def test_raw_socket_sendto_recvfrom_into(self, port: int) -> None:
+        address = ("127.0.0.1", port)
+        secret = b"a very secret message"
+        buffer = bytearray(b"\0" * 256)
+        with ClientContext(
+            DTLSConfiguration(
+                pre_shared_key=("cli", b"secret"),
+                validate_certificates=False,
+            ),
+        ).wrap_socket(
+            socket.socket(socket.AF_INET, socket.SOCK_DGRAM), "localhost"
+        ) as client:
+            client.do_handshake(address)
+            sent = client.sendto(secret, address)
+            assert sent == len(secret)
+
+            received, addr = client.recvfrom_into(buffer, 1024)
+            assert addr == address
+        assert buffer[:received] == secret
